/**********************************************************************
   Mark Butcher    Bsc (Hons) MPhil MIET

   M.J.Butcher Consulting
   Birchstrasse 20f,    CH-5406, Rtihof
   Switzerland

   www.uTasker.com    Skype: M_J_Butcher

   ---------------------------------------------------------------------
   File:        config.h
   Project:     Single Chip Embedded Internet - boot loader
   ---------------------------------------------------------------------
   Copyright (C) M.J.Butcher Consulting 2004..2012
   *********************************************************************

   11.08.2007 Add M5223X SPI FLASH support                               {1}
   16.12.2007 Add SAM7X SPI FLASH support                                {2}
   16.12.2007 FLASH and SRAM size set up added                           {3}
   16.12.2007 Add BOOT_LOADER define                                     {4}
   16.12.2007 Add SPI_CS0_PORT define                                    {5}
   02.04.2008 Add LPC23XX support                                        {6}
   02.04.2008 Add new M5223X clock and PLL details                       {7}
   04.04.2008 Add LM3SXXXX support                                       {8}
   31.05.2008 Clean out all non-bootloader defines (M5223X)
   26.06.2008 Add ST SPI FLASH support                                   {9}
   19.08.2008 Add M5222X and M5221X support (USB boot loader)            {10}
   07.01.2009 Change setting to distinguish between M5221x and M5222x    {11}
   17.01.2009 Add various LM3S derivatives                               {12}
   17.01.2009 Add LM3S SPI FLASH support                                 {13}
   19.01.2009 Add M5225X support                                         {14}
   25.01.2009 Add SST SPI FLASH                                          {15}
   07.06.2009 Add AVR32 support                                          {16}
   04.04.2010 Add LPC17XX support                                        {17}
   30.07.2010 Add dummy SET_SPI_FLASH_MODE() and REMOVE_SPI_FLASH_MODE() for compatibility and correct CONFIGURE_CS_LINES() for Luminary {18}
   29.11.2010 Add RX6XX support                                          {19}
   03.03.2011 Correct SPI_FLASH_AT45DB161, SPI_FLASH_AT45DB321 and SPI_FLASH_AT45DB642 sizes
   29.03.2011 Add Kinetis support                                        {20}
   17.02.2012 Add STM32 support                                          {21}

*/

#ifndef __CONFIG__
#define __CONFIG__

#define MY_PROJECT_NAME     "uTasker Bootloader"

#define BOOT_LOADER                                                      // {4}

#define uDisable_Interrupt  __disable_interrupt
#define uEnable_Interrupt   __enable_interrupt


//#define SPI_SW_UPLOAD                                                  // new SW is situated in SPI FLASH {1}{2}
//#define SPI_FLASH_SST25                                                // {15} use SST SPI FLASH rather than ATMEL
//#define SPI_FLASH_ST                                                   // define that we are using ST FLASH rather than default ATMEL {9}
//#define SPI_DATA_FLASH                                                 // FLASH type is data FLASH supporting sub-sectors (relevant for ST types) {9}

//#define KIRIN3                                                         // M5225X family - otherwise remove!

// Specific project builds
//#define MJB_BUILD_BC2

#define SET_SPI_FLASH_MODE()                                             // {18}
#define REMOVE_SPI_FLASH_MODE()

#ifndef _BOOT_LOADER_ASSEMBLER
    #ifdef _CODE_WARRIOR_CF
        #pragma const_strings on                                         // ensure strings are of const type
    #endif

#ifdef _STM32
  //#define STM3210C_EVAL                                                // evaluation board with STM32F107VCT
    #define STM3240G_EVAL                                                // evaluation board with STM32F407IGH6
  //#define ST_MB913C_DISCOVERY                                          // discovery board with STM32F100RB
  //#define ST_MB997A_DISCOVERY                                          // discovery board with STM32F407VGT6
    #if defined STM3210C_EVAL
        #define TARGET_HW       "STM3210C-EVAL (STM32F107VCT)"
    #elif defined STM3240G_EVAL
        #define TARGET_HW       "STM3240C-EVAL (STM32F407IGH6)"
    #elif defined ST_MB997A_DISCOVERY
        #define TARGET_HW       "MB997A DISCOVERY (STM32F407VGT6)"
    #elif defined ST_MB913C_DISCOVERY
        #define TARGET_HW       "MB913C DISCOVERY (STM32F100RBT6B)"
    #endif
#endif

    #define OUR_HEAP_SIZE (unsigned short)((25*1024))                    // for simulator compatibility only
    #define TICK_RESOLUTION        50                                    // for simulator compatibility only

    #if defined STM3210C_EVAL                                            // STM32F107VCT (72MHz)
        #define CRYSTAL_FREQ        25000000
      //#define DISABLE_PLL                                              // run from clock source directly
      //#define USE_HSI_CLOCK                                            // use internal HSI clock source
        #define USE_PLL2_CLOCK                                           // use the PLL2 output as PLL input (don't use USE_HSI_CLOCK in this configuration)
        #define PLL2_INPUT_DIV      5                                    // clock input is divided by 5 to give 5MHz to the PLL2 input (range 1..16)
        #define PLL2_VCO_MUL        8                                    // the pll2 frequency is multiplied by 8 to 40MHz (range 8..14 or 16 or 20)
        #define PLL_INPUT_DIV       5                                    // 1..16 - should set the input to pll in the range 3..12MHz - not valid for HSI clock source
        #define PLL_VCO_MUL         9                                    // 4..9 where PLL out must be 18..72MHz. Also 65 is accepted as x6.5 (special case)
        #define PIN_COUNT           PIN_COUNT_100_PIN
        #define PACKAGE_TYPE        PACKAGE_LQFP
        #define _STM32F107X                                              // part group
        #define SIZE_OF_RAM         (64 * 1024)                          // 64k SRAM
        #define SIZE_OF_FLASH       (256 * 1024)                         // 256k FLASH
        #define PCLK1_DIVIDE        2
        #define PCLK2_DIVIDE        1
    #elif defined STM3240G_EVAL                                          // STM32F407IGH6 (168MHz)
        #define CRYSTAL_FREQ        25000000
      //#define DISABLE_PLL                                              // run from clock source directly
      //#define USE_HSI_CLOCK                                            // use internal HSI clock source
        #define PLL_INPUT_DIV       25                                   // 2..64 - should set the input to pll in the range 1..2MHz (with preference near to 2MHz)
        #define PLL_VCO_MUL         336                                  // 64 ..432 where VCO must be 64..432MHz
        #define PLL_POST_DIVIDE     2                                    // post divide VCO by 2, 4, 6, or 8 to get the system clock speed
        #define PIN_COUNT           PIN_COUNT_176_PIN
        #define PACKAGE_TYPE        PACKAGE_BGA
        #define _STM32F4XX
        #define SIZE_OF_RAM         (128 * 1024)                         // 128k SRAM
        #define SIZE_OF_CCM         (64 * 1024)                          // 64k Core Coupled Memory
        #define SIZE_OF_FLASH       (1024 * 1024)                        // 1M FLASH
        #define SUPPLY_VOLTAGE      SUPPLY_2_7__3_6                      // power supply is in the range 2.7V..3.6V
        #define PCLK1_DIVIDE        4
        #define PCLK2_DIVIDE        2
        #define HCLK_DIVIDE         1
    #elif defined ST_MB997A_DISCOVERY                                    // STM32F407VGT6 (168MHz)
        #define CRYSTAL_FREQ        8000000
      //#define DISABLE_PLL                                              // run from clock source directly
      //#define USE_HSI_CLOCK                                            // use internal HSI clock source
        #define PLL_INPUT_DIV       4                                    // 2..64 - should set the input to pll in the range 1..2MHz (with preference near to 2MHz)
        #define PLL_VCO_MUL         168                                  // 64 ..432 where VCO must be 64..432MHz
        #define PLL_POST_DIVIDE     2                                    // post divide VCO by 2, 4, 6, or 8 to get the system clock speed
        #define PIN_COUNT           PIN_COUNT_100_PIN
        #define PACKAGE_TYPE        PACKAGE_LQFP
        #define _STM32F4XX
        #define SIZE_OF_RAM         (128 * 1024)                         // 128k SRAM
        #define SIZE_OF_CCM         (64 * 1024)                          // 64k Core Coupled Memory
        #define SIZE_OF_FLASH       (1024 * 1024)                        // 1M FLASH
        #define SUPPLY_VOLTAGE      SUPPLY_2_7__3_6                      // power supply is in the range 2.7V..3.6V
        #define PCLK1_DIVIDE        4
        #define PCLK2_DIVIDE        2
        #define HCLK_DIVIDE         1
    #elif defined ST_MB913C_DISCOVERY                                    // STM32F100RB (24MHz)
        #define CRYSTAL_FREQ        8000000
      //#define DISABLE_PLL                                              // run from clock source directly
      //#define USE_HSI_CLOCK                                            // use internal HSI clock source
        #define PLL_INPUT_DIV       2                                    // 1..16 - should set the input to pll in the range 1..24MHz (with preference near to 8MHz) - not valid for HSI clock source
        #define PLL_VCO_MUL         6                                    // 2 ..16 where PLL out must be 16..24MHz
        #define _STM32F100X                                              // part group
        #define SIZE_OF_RAM         (8 * 1024)                           // 8k SRAM
        #define SIZE_OF_FLASH       (128 * 1024)                         // 128k FLASH
        #define PIN_COUNT           PIN_COUNT_64_PIN
        #define PACKAGE_TYPE        PACKAGE_LQFP
        #define STM32F100RB                                              // exact processor type
        #define PCLK1_DIVIDE        2
        #define PCLK2_DIVIDE        1
    #else
                                                                         // other configurations can be added here
    #endif


#ifdef DISABLE_PLL
    #undef PLL_OUTPUT_FREQ
    #ifdef USE_HSI_CLOCK
        #define PLL_OUTPUT_FREQ  HSI_FREQUENCY
    #else
        #define PLL_OUTPUT_FREQ  CRYSTAL_FREQ
    #endif
#endif
#define SYSCLK          PLL_OUTPUT_FREQ 

#ifdef SPI_FLASH_ST                                                      // ST SPI FLASH used
  //#define SPI_FLASH_STM25P05                                           // the available ST chips
  //#define SPI_FLASH_STM25P10
  //#define SPI_FLASH_STM25P20
  //#define SPI_FLASH_STM25P40
  //#define SPI_FLASH_STM25P80                                               
  //#define SPI_FLASH_STM25P16                                               
  //#define SPI_FLASH_STM25P32                                               
    #define SPI_FLASH_STM25P64
  //#define SPI_FLASH_STM25P128

    #if defined SPI_FLASH_STM25P05                                       // 64k
        #define SPI_FLASH_PAGES             (1*256)                          
    #elif defined SPI_FLASH_STM25P10                                     // 128k
        #define SPI_FLASH_PAGES             (2*256)
    #elif defined SPI_FLASH_STM25P20                                     // 256k
        #define SPI_FLASH_PAGES             (4*256)
    #elif defined SPI_FLASH_STM25P40                                     // 512k
        #define SPI_FLASH_PAGES             (8*256)
    #elif defined SPI_FLASH_STM25P80                                     // 1M
        #define SPI_FLASH_PAGES             (16*256)
    #elif defined SPI_FLASH_STM25P16                                     // 2M
        #define SPI_FLASH_PAGES             (32*256)
    #elif defined SPI_FLASH_STM25P32                                     // 4M
        #define SPI_FLASH_PAGES             (64*256)
    #elif defined SPI_FLASH_STM25P64                                     // 8M
        #define SPI_FLASH_PAGES             (128*256)
    #elif defined SPI_FLASH_STM25P128                                    // 16M
        #define SPI_FLASH_PAGES             (256*256)
    #endif

    #define SPI_FLASH_PAGE_LENGTH 256
    #ifdef SPI_DATA_FLASH
        #define SPI_FLASH_SECTOR_LENGTH (16*SPI_FLASH_PAGE_LENGTH)       // sub-sector size of data FLASH
    #else
        #define SPI_FLASH_SECTOR_LENGTH (256*SPI_FLASH_PAGE_LENGTH)      // sector size of code FLASH
    #endif
    #define SPI_FLASH_BLOCK_LENGTH  SPI_FLASH_SECTOR_LENGTH
#elif defined SPI_FLASH_SST25
  //#define SPI_FLASH_SST25VF010A                                        // the supported SST chips
  //#define SPI_FLASH_SST25LF020A
  //#define SPI_FLASH_SST25LF040A
  //#define SPI_FLASH_SST25VF040B
  //#define SPI_FLASH_SST25VF080B
  //#define SPI_FLASH_SST25VF016B
    #define SPI_FLASH_SST25VF032B

    #if defined SPI_FLASH_SST25VF010A                                    // 1Mbit (128k)
        #define SST25_A_VERSION
        #define SPI_FLASH_PAGES             (32)
    #elif defined SPI_FLASH_SST25LF020A                                  // 2Mbit (256k)
        #define SST25_A_VERSION
        #define SPI_FLASH_PAGES             (64)
    #elif defined SPI_FLASH_SST25LF040A                                  // 4Mbit (512k)
        #define SST25_A_VERSION
        #define SPI_FLASH_PAGES             (128)
    #elif defined SPI_FLASH_SST25VF040B                                  // 4Mbit (512k)
        #define SPI_FLASH_PAGES             (128)
    #elif defined SPI_FLASH_SST25VF080B                                  // 8Mbit (1M)
        #define SPI_FLASH_PAGES             (256)
    #elif defined SPI_FLASH_SST25VF016B                                  // 16Mbit (2M)
        #define SPI_FLASH_PAGES             (512)
    #elif defined SPI_FLASH_SST25VF032B                                  // 32Mbit (4M)
        #define SPI_FLASH_PAGES             (1024)
    #endif

    #define SPI_FLASH_SUB_SECTOR_LENGTH  (4*1024)                        // sub-sector size of SPI FLASH
    #define SPI_FLASH_HALF_SECTOR_LENGTH (32*1024)                       // half-sector size of SPI FLASH
    #define SPI_FLASH_SECTOR_LENGTH      (64*1024)                       // sector size of SPI FLASH (not available on A-versions)

    #define SPI_FLASH_PAGE_LENGTH        SPI_FLASH_SUB_SECTOR_LENGTH     // for compatibility - smallest erasable block
    #define SPI_FLASH_BLOCK_LENGTH       SPI_FLASH_SUB_SECTOR_LENGTH     // for compatibility - file system granularity
#else
    #define SPI_FLASH_ATMEL                                              // default when no other type specified

  //#define SPI_FLASH_AT45DB011                                          // define the ATMEL type used here
  //#define SPI_FLASH_AT45DB021
  //#define SPI_FLASH_AT45DB041
  //#define SPI_FLASH_AT45DB081
  //#define SPI_FLASH_AT45DB161
    #define SPI_FLASH_AT45DB321
  //#define SPI_FLASH_AT45DB642
    #define SPI_FLASH_AT45XXXXD_TYPE                                     // specify that a D-type rather than a B type is being used

    #if defined SPI_FLASH_AT45DB642                                      // define whether used in power of 2 mode or not
        #define SPI_FLASH_PAGE_LENGTH 1056                               // standard page size (B-device only allows 1056)
      //#define SPI_FLASH_PAGE_LENGTH 1024                               // size when power of 2 mode selected (only possible on D-device)
        #define SPI_FLASH_PAGES             (8*1024)                     // 8Meg part {23}
    #elif defined SPI_FLASH_AT45DB321 || defined SPI_FLASH_AT45DB161
        #define SPI_FLASH_PAGE_LENGTH 528                                // standard page size (B-device only allows 528)
      //#define SPI_FLASH_PAGE_LENGTH 512                                // size when power of 2 mode selected (only possible on D-device)
        #if defined SPI_FLASH_AT45DB161
            #define SPI_FLASH_PAGES         (4*1024)                     // 2Meg part {23}
        #else
            #define SPI_FLASH_PAGES         (8*1024)                     // 4Meg part {23}
        #endif
    #else
        #define SPI_FLASH_PAGE_LENGTH 264                                // standard page size (B-device only allows 264)
      //#define SPI_FLASH_PAGE_LENGTH 256                                // size when power of 2 mode selected (only possible on D-device)
        #if defined SPI_FLASH_AT45DB011
            #define SPI_FLASH_PAGES         (512)                        // 128k part
        #elif defined SPI_FLASH_AT45DB021
            #define SPI_FLASH_PAGES         (1024)                       // 256k part
        #elif defined SPI_FLASH_AT45DB041
            #define SPI_FLASH_PAGES         (2*1024)                     // 512k part
        #elif defined SPI_FLASH_AT45DB081
            #define SPI_FLASH_PAGES         (4*1024)                     // 1Meg part
        #endif
    #endif

    #define SPI_FLASH_BLOCK_LENGTH (8*SPI_FLASH_PAGE_LENGTH)             // block size - a block can be deleted
    #define SPI_FLASH_SECTOR_LENGTH (64*4*SPI_FLASH_PAGE_LENGTH)         // exception sector 0a is 2k and sector 0b is 62k
#endif

#if defined STM3240G_EVAL
    // SPI FLASH system setup
    //
    //#define SPI_FLASH_MULTIPLE_CHIPS                                   // activate when multiple physical chips are used
    #define CS0_LINE                            PORTA_BIT4               // CS0 line used when SPI FLASH is enabled
    #define CS1_LINE                                                     // CS1 line used when extended SPI FLASH is enabled
    #define CS2_LINE                                                     // CS2 line used when extended SPI FLASH is enabled
    #define CS3_LINE                                                     // CS3 line used when extended SPI FLASH is enabled

    #define SPI_CS0_PORT                        GPIOA_ODR                // for simulator
    #define SPI_TX_BYTE                         SPI3_DR                  // for simulator
    #define SPI_RX_BYTE                         SPI3_DR                  // for simulator

    #define SSPDR_X                             SPI3_DR
    #define SSPSR_X                             SPI3_SR

    #define __ASSERT_CS(cs_line)                _CLEARBITS(A, cs_line)
    #define __NEGATE_CS(cs_line)                _SETBITS(A, cs_line)

    // SPI 3 used for SPI Flash interface - speed set to 22MHz (PCLK1 168MHz/4)
    //
    #define POWER_UP_SPI_FLASH_INTERFACE()      POWER_UP(APB1, (RCC_APB1ENR_SPI3EN))
    #define POWER_DOWN_SPI_FLASH_INTERFACE()    POWER_DOWN(APB1, (RCC_APB1ENR_SPI3EN))
    #define CONFIGURE_SPI_FLASH_INTERFACE()     _CONFIG_DRIVE_PORT_OUTPUT_VALUE(A, CS0_LINE, (OUTPUT_FAST | OUTPUT_PUSH_PULL), CS0_LINE); \
    _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_SPI3_I2S3ext), (SPI3_CLK_B_3 | SPI3_MOSI_B_5), (OUTPUT_FAST | OUTPUT_PUSH_PULL | INPUT_PULL_DOWN)); \
    _CONFIG_PERIPHERAL_INPUT(B,  (PERIPHERAL_SPI3_I2S3ext), (SPI3_MISO_B_4), INPUT_PULL_DOWN); \
    _RESET_CYCLE_PERIPHERAL(APB1, RCC_APB1ENR_SPI3RST); \
    SPI3_CR1 = (SPICR1_BR_PCLK2_DIV2 | SPICR1_MSTR | SPICR1_SSI | SPICR1_CPOL | SPICR1_CPHA | SPICR1_SSM); \
    SPI3_I2SCFGR = 0; \
    SPI3_CR1 = (SPICR1_SPE | SPICR1_BR_PCLK2_DIV2 | SPICR1_MSTR | SPICR1_SSI | SPICR1_CPOL | SPICR1_CPHA | SPICR1_SSM)
#endif

#define SPI_DATA_FLASH_0_SIZE       (SPI_FLASH_PAGES*SPI_FLASH_PAGE_LENGTH) 
#define SPI_DATA_FLASH_1_SIZE       SPI_DATA_FLASH_0_SIZE 
#define SPI_DATA_FLASH_2_SIZE       SPI_DATA_FLASH_0_SIZE
#ifdef SPI_FLASH_MULTIPLE_CHIPS
    #define SPI_FLASH_DEVICE_COUNT  3                                    // only define CS_LINE for this amount of devices
    #if SPI_FLASH_DEVICE_COUNT > 2
        #define SPI_DATA_FLASH_SIZE   (SPI_DATA_FLASH_0_SIZE + SPI_DATA_FLASH_1_SIZE + SPI_DATA_FLASH_2_SIZE)
    #else
        #define SPI_DATA_FLASH_SIZE   (SPI_DATA_FLASH_0_SIZE + SPI_DATA_FLASH_1_SIZE)
    #endif
#else
    #define SPI_DATA_FLASH_SIZE       SPI_DATA_FLASH_0_SIZE                                  
#endif


#define SPI_FLASH_START        (FLASH_START_ADDRESS + SIZE_OF_FLASH)     // SPI FLASH starts immediately after FLASH
#define SUPPORT_MIME_IDENTIFIER                                          // set when the application uses this parameter

#define ACTIVE_FILE_SYSTEM                                               // needed for simulating internal flash operation
#if defined SPI_SW_UPLOAD
    #define FILE_GRANULARITY   (1*SPI_FLASH_BLOCK_LENGTH)
    #define uFILE_START        (unsigned char *)(SPI_FLASH_START + (61*FILE_GRANULARITY)) // the address where the new application is located when present - uFileSystem SPI location "z"
    #define FILE_SYSTEM_SIZE   (128 * 1024)                              // maximum size of new application
#else
    #define FILE_GRANULARITY   (1*FLASH_GRANULARITY)
    #define uFILE_START        (unsigned char *)(FLASH_START_ADDRESS + 0x60000) // the address where the new application is located when present
    #define FILE_SYSTEM_SIZE   (128 * 1024)                              // maximum size of new application
#endif



#ifdef SUPPORT_MIME_IDENTIFIER
  #define FILE_HEADER (sizeof(MAX_FILE_LENGTH) + 1)                      // file length followed by MIME identifier
#else
  #define FILE_HEADER (sizeof(MAX_FILE_LENGTH))
#endif


#define PHYSICAL_QUEUES  0
#define CLOCK_LIMIT      DELAY_LIMIT


/**********************************************************************************************************/


#include "types.h"                                                       // project specific type settings
#include "../../uTasker\uTasker.h"
#include "../../uTasker\driver.h"
#include "../../stack\tcpip.h"
#ifdef _WINDOWS
    #include "../../WinSim\WinSim.h"
#endif
#include "../../Hardware/hardware.h"

#ifdef SPI_SW_UPLOAD
    extern int fnConfigSPIFileSystem(void);
#endif


/************ uTasker task table is defined here but only used by the hardware module initiating the system ***********/

#if defined OPSYS_CONFIG && defined _WINDOWS                             // this is only set in the hardware module
    #define DEFAULT_NODE_NUMBER 1
    CONFIG_LIMIT OurConfigNr = DEFAULT_NODE_NUMBER;                      // in single node system this can be initialised with a fixed value

    const HEAP_NEEDS ctOurHeap[] = {
        {DEFAULT_NODE_NUMBER, OUR_HEAP_SIZE},                            // our node requires this amount of heap space
        {0}                                                              // end
    };


    const UTASK_TASK ctNodes[] = {                                       // we use a single fixed configuration (single node)
        DEFAULT_NODE_NUMBER,                                             // configuration our single node
        0,                                                               // end of single configuration

        // insert more node configurations here if required
        0                                                                // end of configuration list
    };


    const UTASKTABLEINIT ctTaskTable[] = {
        0
    };
#endif                                                                   // end of task configuration




extern unsigned short fnCRC16(unsigned short usCRC, unsigned char *ptrInput, unsigned short usBlockSize);
#ifdef _COMPILE_KEIL
    extern int uTaskerBoot(void);
#else
    extern void uTaskerBoot(void);
#endif
extern unsigned long fnEraseFlash(unsigned char *ptrSector);
extern void fnGetPars(unsigned char *ParLocation, unsigned char *ptrValue, MAX_FILE_LENGTH Size);

#endif
#endif
