/***********************************************************************
    Mark Butcher    Bsc (Hons) MPhil MIET

    M.J.Butcher Consulting
    Birchstrasse 20f,    CH-5406, Rtihof
    Switzerland

    www.uTasker.com    Skype: M_J_Butcher

    ---------------------------------------------------------------------
    File:      Loader.h
    Project:   uTasker Demonstration project
    ---------------------------------------------------------------------
    Copyright (C) M.J.Butcher Consulting 2004..2014
    *********************************************************************
    15.09.2009 Add STR91XF setup
    03.01.2010 Add SAM7X setup
    26.03.2010 LPC2XXX application start set to 8k and add intermediate buffer {1}
    05.10.2010 Add optional SHOW_APP_DETAILS                             {2}
    01.12.2010 Add RX6XX                                                 {3}
    02.04.0211 Add Kinetis                                               {4}
    14.05.2011 Add USB for Coldfire                                      {5}
    31.05.2011 Add USB for SAM7X                                         {6}
    03.06.2011 Activate USB_MSD password protection for M522xx and Kinetis {7}
    22.01.2012 Add Kinetis SD card loader configuration                  {8}
    11.02.2012 Add M522xx SD card loader configuration                   {9}
    12.02.2012 Add STM32                                                 {10}
    18.02.2012 Add LPC2xxx SD card configuration                         {11}
    20.04.2012 Add TWR_K20N50M                                           {12}
    03.03.2013 Add fnAddSREC_file()                                      {13}
    25.11.2013 Add MAX_WAIT_SD_CARD                                      {14}
    16.01.2014 Add FRDM_K20D50M, FRDM_KL46Z and TWR_K70F120M             {15}
    28.01.2014 Add TWR_KL46Z48M                                          {16}
    29.01.2014 Add FRDM_KL25Z, FRDM_KL26Z and TWR_KL25Z48M               {17}
    01.02.2014 Add FRDM_KL02Z, FRDM_KL05Z and FRDM_KE02Z                 {18}
    15.04.2014 Add FRDM_K64F and TWR_K64F120M                            {19}
    15.04.2014 Add web server based software upload support              {20}
    15.05.2014 Add optional SD card wildcard file name matching          {21}
    15.05.2014 Add optional decryption of SD card content                {22}

*/

#define SOFTWARE_VERSION              "V1.1"

#define MY_PROJECT_NAME               "uTasker loader project"

#define MAX_WAIT_SD_CARD              4                                  // {14} if no SD card is not detected after 4s the existing application is started

//#define WILDCARD_FILES                                                 // {21} allow wildcard name matching to be used
#if defined WILDCARD_FILES
    #undef _REMOVE_FORMATTED_OUTPUT
#endif
//#define ENCRYPTED_CARD_CONTENT                                         // {22}

#define LOADER_UART                   0                                  // the serial interface used by the serial loader

#define SHOW_APP_DETAILS                                                 // {2} display application start and size in menu
#define SERIAL_INTERFACE_MODE         (CHAR_8 + NO_PARITY + ONE_STOP + USE_XON_OFF + CHAR_MODE) // the UART configuration


#if defined _STM32                                                       // {10}
    #define SERIAL_SPEED              SERIAL_BAUD_115200                 // the Baud rate of the UART
    #define ROOT_FILE_ENTRIES         4                                  // when USB MSD loader this many directory entries are set to the start of FLASH - the application start is shifted by this amount x 32 bytes
    #define ENABLE_READBACK                                              // allow USB to transfer present application to PC
    #define UTASKER_APP_START         (FLASH_START_ADDRESS + (16 * 1024))// application starts at this address
    #define UTASKER_APP_END           (unsigned char *)(UTASKER_APP_START + (130 * 1024)) // end of application space - after maximum application size

    // Before software can be read from the disk a password file must have been copied
    //
    #define READ_PASSWORD             "enable file read from the Kinetis device by dragging this file to the disk" // password with maximum length of 512 bytes

    // SD card loading - file to be loaded, magic number and secret key for authenticating the file's content {8}
    //
    #if defined ENCRYPTED_CARD_CONTENT
        #define NEW_SOFTWARE_FILE     "sd_card_enc.bin"
        #define VALID_VERSION_MAGIC_NUMBER   0x651d
        #define _SECRET_KEY           {0xb7, 0x48, 0xb6, 0x53, 0x11, 0x24}
        static const unsigned char ucDecrypt[] = {0xff, 0x25, 0xa7, 0x88, 0xf2, 0xe6, 0x81, 0x33, 0x87, 0x77}; // must be even in length (dividable by unsigned short)
        #define KEY_PRIME             0xafe1                             // never set to 0
        #define CODE_OFFSET           0xc298                             // ensure that this value is a multiple of the smallest flash programming entity size (divisible by 8 is suitable for all Kinetis parts)
    #else
        #define NEW_SOFTWARE_FILE            "software.bin"
        #define VALID_VERSION_MAGIC_NUMBER   0x651c
        #define _SECRET_KEY                  {0x9c, 0x66, 0x9a, 0x00, 0x8f, 0x71}
    #endif
#endif

#if defined USB_INTERFACE
    #define _UTASKER_APP_START_ (UTASKER_APP_START + (ROOT_FILE_ENTRIES * 32)) // when USB is used the start of application space is used for FAT entries
#else
    #define _UTASKER_APP_START_ (UTASKER_APP_START)
#endif

#if !defined UPLOAD_OFFSET
    #define UPLOAD_OFFSET 0
#endif

#define POINTER_USER_NAME "ADMIN"                                        // {20} optional web server authentication
#define POINTER_USER_PASS "admin"

#define MAX_UTASKER_APP_SIZE      (MAX_FILE_LENGTH)(UTASKER_APP_END - (unsigned char *)UTASKER_APP_START)



extern void fnConfigureAndStartWebServer(void);

#define T_RESET                   1                                      // application timer events
#define T_GO_TO_APP               2
#define T_COMMIT_BUFFER           3

extern int fnAddSREC_file(unsigned char *ptrLastAddress);                // {13}

