/**********************************************************************
    Mark Butcher    Bsc (Hons) MPhil MIET

    M.J.Butcher Consulting
    Birchstrasse 20f,    CH-5406, Rtihof
    Switzerland

    www.uTasker.com    Skype: M_J_Butcher

    ---------------------------------------------------------------------
    File:      webInterface.c
    Project:   Ethernet Boot Loader
    ---------------------------------------------------------------------
    Copyright (C) M.J.Butcher Consulting 2004..2014
    *********************************************************************

*/


/* =================================================================== */
/*                           include files                             */
/* =================================================================== */

#include "config.h"

/* =================================================================== */
/*                             constants                               */
/* =================================================================== */

#if defined USE_HTTP

const CHAR  cSoftwareVersion[] = SOFTWARE_VERSION;                       // software version for general purpose display use
const CHAR  cProcessorName[]   = TARGET_HW;                              // processor/board name for display use

/* =================================================================== */
/*                      local structure definitions                    */
/* =================================================================== */

/* =================================================================== */
/*                          local definitions                          */
/* =================================================================== */

/* =================================================================== */
/*                 local function prototype declarations               */
/* =================================================================== */

static void fnDelayResetBoard(void);
static int  fnCheckErasePassword(int iType, CHAR *ptrData);

/* =================================================================== */
/*                     global variable definitions                     */
/* =================================================================== */

NETWORK_PARAMETERS network[IP_NETWORK_COUNT] = {                         // fixed network values
    {
    (AUTO_NEGOTIATE /*| FULL_DUPLEX*/ | RX_FLOW_CONTROL),                // usNetworkOptions - see driver.h for other possibilities
    {0x00, 0x00, 0x00, 0x00, 0x00, 0x00},                                // ucOurMAC - when no other value can be read from parameters this will be used
    { 192, 168, 0, 125 },                                                // ucOurIP - our default IP address
    { 255, 255, 255, 0 },                                                // ucNetMask - Our default network mask
    { 192, 168, 0, 1 },                                                  // ucDefGW - Our default gateway
    { 192, 168, 0, 1 },                                                  // ucDNS_server - Our default DNS server
#if defined USE_IPV6
    { _IP6_ADD_DIGIT(0x2001), _IP6_ADD_DIGIT(0x0470), _IP6_ADD_DIGIT(0x0026), _IP6_ADD_DIGIT(0x0105), _IP6_ADD_DIGIT(0x0000), _IP6_ADD_DIGIT(0x0000), _IP6_ADD_DIGIT(0x0000), _IP6_ADD_DIGIT(0x0030) }, // default global IPV6 address
#endif
    },
};

/* =================================================================== */
/*                             constants                               */
/* =================================================================== */

const CHAR  cNoSW[] = "No Application SW present";
const CHAR  cSW_detected[] = "Software Detected";

// Erase password
// p12X-k3ve2B1O2Ba
//
const CHAR  cErasePass[] = {'p', '1', '2', 'X', '-', 'k', '3', 'v', 'e', '2', 'B', '1', 'O', '2', 'B', 'a'};

// Mass erase password
// mMm122-aHHHQq1x8
//
const CHAR  cMassErasePass[] = {'m', 'M', 'm', '1', '2', '2', '-', 'a', 'H', 'H', 'H', 'Q', 'q', '1', 'x', '8'};
const CHAR  cEmptyPass[]     = {'*', '*', '*', '*', '*', '*', '*', '*'};



/* =================================================================== */
/*                      local variable definitions                     */
/* =================================================================== */


static unsigned char  fnIsSelected(unsigned char *ptrBuffer);
#ifdef HTTP_DYNAMIC_CONTENT
    static CHAR      *fnInsertString(unsigned char *ptrBuffer, LENGTH_CHUNK_COUNT TxLength, unsigned short *usLengthToSend, HTTP *http_session);
#else
    static CHAR      *fnInsertString(unsigned char *ptrBuffer, LENGTH_CHUNK_COUNT TxLength, unsigned short *usLengthToSend);
#endif
static int            fnHandleWeb(unsigned char ucType, CHAR *ptrData, HTTP *http_session);

#define UPLOAD_FAILED         "<html><head><title>SW Upload failed</title></head><body bgcolor=#ff9000 text=#000000 topmargin=3 marginheight=3><center><td valign=top class=h><font color=#ff0000 style=font-size:30px><b style='mso-bidi-font-weight:normal'>&micro;Tasker</font></i></b></td><br></td><td align=left><br><br>Sorry but upload has failed.</font><br><br><a href=""javascript:history.back()"">Return</a></body></html>"
#define SW_UPLOAD_COMPLETED   "<html><head><meta http-equiv=""refresh"" content=""5;URL=0Menu.htm""><title>KINETIS SW Uploaded</title></head><body bgcolor=#ffffff text=#000000 topmargin=3 marginheight=3><center><td valign=top class=h><font color=#ff0000 style=font-size:30px><b style='mso-bidi-font-weight:normal'>&micro;Tasker</font> - SW Update</i></b></td><br></td><td align=left><br><br>SW Upload successful. The KINETIS target will now reset and start the new program. Please wait 5s...</body></html>"


static const CHAR cWarningHTML[]       = UPLOAD_FAILED;
static const CHAR cSuccessSW_HTML[]    = SW_UPLOAD_COMPLETED;


extern void fnConfigureAndStartWebServer(void)
{
    #if defined AUTHENTICATE_WEB_ACCESS
    unsigned char ucHttpProperties = (WEB_SUPPORT_PARAM_GEN | WEB_SUPPORT_HANDLER | WEB_AUTHENTICATE);
    #else
    unsigned char ucHttpProperties = (WEB_SUPPORT_PARAM_GEN | WEB_SUPPORT_HANDLER);
    #endif
    #ifdef _VARIABLE_HTTP_PORT
    fnStartHTTP(fnHandleWeb, fnIsSelected, fnInsertString, ucHttpProperties, HTTP_SERVERPORT); // {27} pass the port number to be used, which can be variable of required
    #else
    fnStartHTTP(fnHandleWeb, fnIsSelected, fnInsertString, ucHttpProperties);
    #endif
}


// We can define whether a selection is set or not when a web page is displayed
//
static unsigned char fnIsSelected(unsigned char *ptrBuffer)
{
    int iDisabled = 0;
    // Format sAB (length assumed to be 4 bytes, even when both parameters are not used)
    //
    switch (*(ptrBuffer + 1)) {
    case 's':
        if (*(unsigned long *)fnGetFlashAdd((unsigned char *)_UTASKER_APP_START_) == 0xffffffff) { // check whether there is code at the start address
            iDisabled = 1;
        }
        if (*ptrBuffer == WEB_NOT_DISABLE_FIELD) {                       // if the condition is to be inverted
            iDisabled = !iDisabled;
        }
        if (iDisabled != 0) {
            return IS_DISABLED;
        }
        break;

    default:
        break;

    }
    return NOT_SELECTED;
}

// Here we handle incoming web parameters - we can also change the file to be displayed if we want to
//
static int fnHandleWeb(unsigned char ucType, CHAR *ptrData, HTTP *http_session)
{
    static unsigned char *ptrCode = 0;
    #if defined FLASH_ROW_SIZE && FLASH_ROW_SIZE > 0
        static unsigned char ucCodeStart[FLASH_ROW_SIZE];
    #else
        static unsigned char ucCodeStart[4] = {0xff, 0xff, 0xff, 0xff};
    #endif
    switch (ucType) {
    case 0:                                                              // this is called on each new GET - we can reset variable which collect data here if required
        break;

    case CAN_POST_BEGIN:
        if (http_session->FileLength < MAX_UTASKER_APP_SIZE) { 
            ptrCode = (unsigned char *)_UTASKER_APP_START_;              // note that an upload is starting
        }
        return (0);                                                      // allow post at all times

    case POSTING_DATA_TO_APP:
        if (ptrCode != 0) {
            unsigned short usBuff_length = (unsigned short)http_session->FileLength;
            if (ptrCode == (unsigned char *)_UTASKER_APP_START_) {       // first frame being received
                if (http_session->FileLength < sizeof(ucCodeStart)) {
                    ptrCode = 0;
                    break;
                }
                uMemcpy(ucCodeStart, ptrData, sizeof(ucCodeStart));
                ptrData += sizeof(ucCodeStart);
                ptrCode += sizeof(ucCodeStart);
                usBuff_length -= sizeof(ucCodeStart);
            }
            fnWriteBytesFlash(ptrCode, (unsigned char *)ptrData, usBuff_length); // program the frame to flash
            ptrCode += usBuff_length;
        }
        break;

    case INFORM_POST_SUCCESS:
#if defined SUPPORT_MIME_IDENTIFIER
        http_session->ucMimeType = MIME_HTML;                            // force HTML type
#endif
        if (http_session->cDisplayFile == 'S') {                         // software upload
            if (ptrCode != 0) {
                fnWriteBytesFlash((unsigned char *)_UTASKER_APP_START_, ucCodeStart, sizeof(ucCodeStart));
#if defined FLASH_ROW_SIZE && FLASH_ROW_SIZE > 0
                fnWriteBytesFlash(0, 0, 0);                              // close any outstanding FLASH buffer
#endif
                http_session->ptrFileStart = (unsigned char*)cSuccessSW_HTML;
                http_session->FileLength = (sizeof(cSuccessSW_HTML) - 1);
                fnDelayResetBoard();                                     // after 2 seconds the application will reset and then the new software will be copied
            }
            else {
                http_session->ptrFileStart = (unsigned char*)cWarningHTML;
                http_session->FileLength   = (sizeof(cWarningHTML) - 1);
            }
            return DISPLAY_INTERNAL;
        }
        break;

    case 'c':                                                            // erase the complete application space
        if (*ptrData++ == '1') {                                         // if erase software
            if (fnCheckErasePassword(1, ++ptrData) != 0) {
                fnEraseFlashSector((unsigned char *)UTASKER_APP_START, (unsigned long)(UTASKER_APP_END - UTASKER_APP_START));
            }
        }
        else {
            if (fnCheckErasePassword(0, ++ptrData) != 0) {
#if defined MASS_ERASE
                fnMassEraseFlash();
#endif
            }
        }
        break;

    default:                                                             // page refreshes should be performed using 'r', which is then not processed
        break;
    }
    return 0;                                                            // OK to display the requested file
}

// We can add strings to a web page being transmitted. If the value usTxLength is not zero we are expected to generate dynamic HTML.
//
#ifdef HTTP_DYNAMIC_CONTENT                                              // {6}{8}
static CHAR *fnInsertString(unsigned char *ptrBuffer, LENGTH_CHUNK_COUNT TxLength, unsigned short *usLengthToSend, HTTP *http_session)
#else
static CHAR *fnInsertString(unsigned char *ptrBuffer, LENGTH_CHUNK_COUNT TxLength, unsigned short *usLengthToSend)
#endif
{
    // vXY commands - to display a value, string etc.
    //
#if !defined WEB_ESCAPE_LEN
    #define WEB_ESCAPE_LEN 4
#endif

    if (!TxLength) {                                                     // this means that we have a simple string field to be filled
        switch (*ptrBuffer++) {
        case 'V':                                                        // display version
            if (*ptrBuffer == '1') {                                     // display application version
                if (*(unsigned long *)fnGetFlashAdd((unsigned char *)_UTASKER_APP_START_) != 0xffffffff) {
                    *usLengthToSend = (sizeof(cSW_detected) - 1);
                    return (CHAR *)cSW_detected;                         // display that SW is present
                }
                else {
                    *usLengthToSend = (sizeof(cNoSW) - 1);
                    return (CHAR *)cNoSW;                                // display that there is no SW loaded
                }
            }
            else {
                *usLengthToSend = (sizeof(cSoftwareVersion) - 1);
                return (CHAR *)cSoftwareVersion;                         // display loader version
            }
            break;

        case 'e':                                                        // erase password - display as empty string
            *usLengthToSend = (sizeof(cEmptyPass));
            return (CHAR *)cEmptyPass;

        case 'N':                                                        // insert the name of the board/processor
            *usLengthToSend = (sizeof(cProcessorName) - 1);
            return (CHAR *)cProcessorName;

        default:
            break;
        }
    }
    *usLengthToSend = 0;
    return 0;
}


static int fnCheckErasePassword(int iType, CHAR *ptrData)
{
    if (iType != 0) {
        if (!uMemcmp(ptrData, cErasePass, sizeof(cErasePass))) {
            return 1;
        }
    }
    else {
        if (!uMemcmp(ptrData, cMassErasePass, sizeof(cMassErasePass))) {
            return 1;
        }
    }
    return 0;
}

// Reset board after a delay to allow present web page to complete serving
//
static void fnDelayResetBoard(void)
{
    uTaskerMonoTimer(TASK_APPLICATION, (DELAY_LIMIT)(2 * SEC), T_RESET);
}

// Dummy function
//
extern QUEUE_TRANSFER fnNetworkTx(unsigned char *output_buffer, QUEUE_TRANSFER nr_of_bytes)
{
    return 0;
}

// Dummy function
//
extern void fnGetEthernetPars(void)
{
}

#endif
