/**********************************************************************
    Mark Butcher    Bsc (Hons) MPhil MIET

    M.J.Butcher Consulting
    Birchstrasse 20f,    CH-5406, Rtihof
    Switzerland

    www.uTasker.com    Skype: M_J_Butcher

    ---------------------------------------------------------------------
    File:      STM32F100X_port.h
    Project:   Single Chip Embedded Internet
    ---------------------------------------------------------------------
    Copyright (C) M.J.Butcher Consulting 2004..2014
    *********************************************************************/

#if PIN_COUNT == PIN_COUNT_100_PIN
    #define _PIN_COUNT              0
#elif PIN_COUNT == PIN_COUNT_64_PIN
    #if PACKAGE_TYPE == PACKAGE_LQFP
        #define _PIN_COUNT          1
    #else
        #define _PIN_COUNT          2
    #endif
#elif defined PIN_COUNT_48_PIN
    #define _PIN_COUNT              3
#else
    #error "Package not clearly defined!!"
#endif

#define ALTERNATIVE_FUNCTIONS   6                                        // GPIO plus 6 possible peripheral functions


static const char *cPer[PORTS_AVAILABLE][PORT_WIDTH][1 + ALTERNATIVE_FUNCTIONS] = {
    {
        //                                                               GPIO A
        {  "PA0", "-",              "-",           "-",          "USART2_CTS", "TIM2_CH1_ETR","ADC1_IN0" }, // PA0
        {  "PA1", "-",              "-",           "TIM2_CH2",   "USART2_RTS", "-",        "ADC1_IN1"  }, // PA1
        {  "PA2", "-",              "-",           "TIM2_CH3",   "USART2_TX",  "TIM15_CH1","ADC1_IN2"  }, // PA2
        {  "PA3", "-",              "-",           "TIM2_CH4",   "USART2_RX",  "TIM15_CH2","ADC1_IN3"  }, // PA3
        {  "PA4", "SPI1_NSS",       "-",           "-",          "USART2_CK",  "DAC1_OUT", "ADC1_IN4"  }, // PA4
        {  "PA5", "SPI1_SCK",       "-",           "-",          "-",          "DAC2_OUT", "ADC1_IN5"  }, // PA5
        {  "PA6", "SPI1_MISO",      "TIM3_CH1",    "TIM1_BKIN",  "-",          "TIM16_CH1","ADC1_IN6"  }, // PA6
        {  "PA7", "SPI1_MOSI",      "TIM3_CH2",    "TIM1_CH1N",  "-",          "TIM17_CH1","ADC1_IN7"  }, // PA7
        {  "PA8", "MCO",            "TIM1_CH1",    "-",          "USART1_CK",  "-",        "-"         }, // PA8
        {  "PA9", "TIM1_CH2",       "USART1_TX",   "-",          "-",          "TIM15_BKIN","-"        }, // PA9
        {  "PA10","TIM1_CH3",       "USART1_RX",   "-",          "-",          "TIM17_BKIN","-"        }, // PA10
        {  "PA11","TIM1_CH4",       "USART1_CTS",  "-",          "-",          "-",        "-"         }, // PA11
        {  "PA12","TIM1_ETR",       "USART1_RTS",  "-",          "-",          "-",        "-"         }, // PA12
        {  "PA13","-",              "-",           "-",          "JTMS-SWDIO", "-",        "-"         }, // PA13
        {  "PA14","-",              "-",           "-",          "JTCK-SWCLK", "-",        "-"         }, // PA14
        {  "PA15","-",              "-",           "TIM2_CH1_ETR","JTDI",      "-",        "-"         }  // PA15
    },
    {
        //                                                               GPIO B
        {  "PB0", "-",              "TIM3_CH3",    "TIM1_CH2N",  "-",          "-",        "ADC1_IN8"  }, // PB0
        {  "PB1", "-",              "TIM3_CH4",    "-",          "-",          "TIM1_CH3N","ADC1_IN9"  }, // PB1
        {  "PB2", "-",              "-",           "-",          "BOOT1",      "-",        "-"         }, // PB2
        {  "PB3", "-",              "-",           "-",          "JTDO",       "TRACESWO", "SPI1_SCK"  }, // PB3
        {  "PB4", "TIM3_CH1",       "SPI1_MISO",   "-",          "NJTRST",     "-",        "-"         }, // PB4
        {  "PB5", "-",              "TIM3_CH2",    "SPI1_MOSI",  "TIM16_BKIN", "I2C1_SMBA","-"         }, // PB5
        {  "PB6", "I2C1_SCL",       "USART1_TX",   "TIM4_CH1",   "TIM16_CH1N", "-",        "-"         }, // PB6
        {  "PB7", "I2C1_SDA",       "USART1_RX",   "TIM4_CH2",   "TIM17_CH1N", "-",        "-"         }, // PB7
        {  "PB8", "-",              "-",           "TIM4_CH3",   "I2C1_SCL",   "TIM16_CH1","CEC"       }, // PB8
        {  "PB9", "-",              "-",           "TIM4_CH4",   "I2C1_SDA",   "TIM17_CH1","-"         }, // PB9
        {  "PB10","-",              "USART3_TX",   "TIM2_CH3",   "I2C2_SCL",   "-",        "CEC"       }, // PB10
        {  "PB11","-",              "USART3_RX",   "TIM2_CH4",   "I2C2_SDA",   "-",        "-"         }, // PB11
        {  "PB12","-",              "USART3_CK",   "TIM1_BKIN",  "SPI2_NSS",   "I2C2_SMBA", "-"        }, // PB12
        {  "PB13","-",              "USART3_CTS",  "TIM1_CH1N",  "SPI2_SCK",   "-",        "-"         }, // PB13
        {  "PB14","SPI2_MISO",      "USART3_RTS",  "TIM1_CH2N",  "-",          "-",        "-"         }, // PB14
        {  "PB15","-",              "-",           "TIM1_CH3N",  "SPI2_MOSI",  "TIM15_CH1N","TIM15_CH2"}  // PB15
    },
    {
        //                                                               GPIO C
        {  "PC0", "-",              "-",           "-",          "-",          "-",        "ADC1_IN10" }, // PC0
        {  "PC1", "-",              "-",           "-",          "-",          "-",        "ADC1_IN11" }, // PC1
        {  "PC2", "-",              "-",           "-",          "-",          "-",        "ADC1_IN12" }, // PC2
        {  "PC3", "-",              "-",           "-",          "-",          "-",        "ADC1_IN13" }, // PC3
        {  "PC4", "-",              "-",           "-",          "-",          "-",        "ADC1_IN14" }, // PC4
        {  "PC5", "-",              "-",           "-",          "-",          "-",        "ADC1_IN15" }, // PC5
        {  "PC6", "TIM3_CH1",       "-",           "-",          "-",          "-",        "-"         }, // PC6
        {  "PC7", "TIM3_CH2",       "-",           "-",          "-",          "-",        "-"         }, // PC7
        {  "PC8", "TIM3_CH3",       "-",           "-",          "-",          "-",        "-"         }, // PC8
        {  "PC9", "TIM3_CH4",       "-",           "-",          "-",          "-",        "-"         }, // PC9
        {  "PC10","USART3_TX",      "-",           "-",          "-",          "-",        "-"         }, // PC10
        {  "PC11","USART3_RX",      "-",           "-",          "-",          "-",        "-"         }, // PC11
        {  "PC12","USART3_CK",      "-",           "-",          "-",          "-",        "-"         }, // PC12
        {  "PC13","-",              "TAMPER-RTC",  "-",          "-",          "-",        "-"         }, // PC13
        {  "PC14","-",              "OSC32_IN",    "-",          "-",          "-",        "-"         }, // PC14
        {  "PC15","-",              "OSC32_OUT",   "-",          "-",          "-",        "-"         }  // PC15
    },
    {
        //                                                               GPIO D
        {  "PD0", "-",              "-",           "OSC_IN",     "CAN1_RX",    "-",        "-"         }, // PD0
        {  "PD1", "-",              "-",           "OSC_OUT",    "CAN1_TX",    "-",        "-"         }, // PD1
        {  "PD2", "UART5_RX",       "TIM3_ETR",    "-",          "-",          "-",        "-"         }, // PD2
        {  "PD3", "-",              "USART2_CTS",  "-",          "-",          "-",        "-"         }, // PD3
        {  "PD4", "-",              "USART2_RTS",  "-",          "-",          "-",        "-"         }, // PD4
        {  "PD5", "-",              "USART2_TX",   "-",          "-",          "-",        "-"         }, // PD5
        {  "PD6", "-",              "USART2_RX",   "-",          "-",          "-",        "-"         }, // PD6
        {  "PD7", "-",              "USART2_CK",   "-",          "-",          "-",        "-"         }, // PD7
        {  "PD8", "ETH_MII_RX_DV",  "USART3_TX",   "-",          "-",          "-",        "-"         }, // PD8
        {  "PD9", "ETH_MII_RXD0",   "USART3_RX",   "-",          "-",          "-",        "-"         }, // PD9
        {  "PD10","ETH_MII_RXD1",   "USART3_CK",   "-",          "-",          "-",        "-"         }, // PD10
        {  "PD11","ETH_MII_RXD2",   "USART3_CTS",  "-",          "-",          "-",        "-"         }, // PD11
        {  "PD12","ETH_MII_RXD3",   "USART3_RTS",  "TIM4_CH1",   "-",          "-",        "-"         }, // PD12
        {  "PD13","-",              "-",           "TIM4_CH2",   "-",          "-",        "-"         }, // PD13
        {  "PD14","-",              "-",           "TIM4_CH3",   "-",          "-",        "-"         }, // PD14
        {  "PD15","-",              "-",           "TIM4_CH4",   "-",          "-",        "-"         }  // PD15
    },
    {
        // GPIO E
        {  "PE0", "TIM4_ETR",       "-",           "-",          "-",          "-",        "-"         }, // PE0
        {  "PE1", "-",              "-",           "-",          "-",          "-",        "-"         }, // PE1
        {  "PE2", "-",              "-",           "-",          "-",          "TRACECK",  "-"         }, // PE2
        {  "PE3", "-",              "-",           "-",          "-",          "TRACED0",  "-"         }, // PE3
        {  "PE4", "-",              "-",           "-",          "-",          "TRACED1",  "-"         }, // PE4
        {  "PE5", "-",              "-",           "-",          "-",          "TRACED2",  "-"         }, // PE5
        {  "PE6", "-",              "-",           "-",          "-",          "TRACED3",  "-"         }, // PE6
        {  "PE7", "TIM1_ETR",       "-",           "-",          "-",          "-",        "-"         }, // PE7
        {  "PE8", "TIM1_CH1N",      "-",           "-",          "-",          "-",        "-"         }, // PE8
        {  "PE9", "TIM1_CH1",       "-",           "-",          "-",          "-",        "-"         }, // PE9
        {  "PE10","TIM1_CH2N",      "-",           "-",          "-",          "-",        "-"         }, // PE10
        {  "PE11","TIM1_CH2",       "-",           "-",          "-",          "-",        "-"         }, // PE11
        {  "PE12","TIM1_CH3N",      "-",           "-",          "-",          "-",        "-"         }, // PE12
        {  "PE13","TIM1_CH3",       "-",           "-",          "-",          "-",        "-"         }, // PE13
        {  "PE14","TIM1_CH4",       "-",           "-",          "-",          "-",        "-"         }, // PE14
        {  "PE15","TIM1_BKIN",      "-",           "-",          "-",          "-",        "-"         }  // PE15
    }
};


static const char *cPinNumber[PORTS_AVAILABLE][PORT_WIDTH][4] = {
    {
        //  LQFP100 LQFP64  BGA64  LQFP48                                GPIO A
        {  "23",    "14",   "G2",  "10"  },                              // PA0
        {  "24",    "15",   "H2",  "11"  },                              // PA1
        {  "25",    "16",   "F3",  "12"  },                              // PA2
        {  "26",    "17",   "G3",  "13"  },                              // PA3
        {  "29",    "20",   "H3",  "14"  },                              // PA4
        {  "30",    "21",   "F4",  "15"  },                              // PA5
        {  "31",    "22",   "G4",  "16"  },                              // PA6
        {  "32",    "23",   "H4",  "17"  },                              // PA7
        {  "67",    "41",   "D7",  "29"  },                              // PA8
        {  "68",    "42",   "C7",  "30"  },                              // PA9
        {  "69",    "43",   "C6",  "31"  },                              // PA10
        {  "70",    "44",   "C8",  "32"  },                              // PA11
        {  "71",    "45",   "B8",  "33"  },                              // PA12
        {  "72",    "46",   "A8",  "34"  },                              // PA13
        {  "76",    "49",   "A7",  "37"  },                              // PA14
        {  "77",    "50",   "A6",  "38"  },                              // PA15
    },

    {
        //  LQFP100 LQFP64  BGA64  LQFP48                                GPIO B
        {  "35",    "26",   "F5",  "18"  },                              // PB0
        {  "36",    "27",   "G5",  "19"  },                              // PB1
        {  "37",    "28",   "G6",  "20"  },                              // PB2
        {  "89",    "55",   "A5",  "39"  },                              // PB3
        {  "90",    "56",   "A4",  "40"  },                              // PB4
        {  "91",    "57",   "C4",  "41"  },                              // PB5
        {  "92",    "58",   "D3",  "42"  },                              // PB6
        {  "93",    "59",   "C3",  "43"  },                              // PB7
        {  "95",    "61",   "B3",  "45"  },                              // PB8
        {  "96",    "62",   "A3",  "46"  },                              // PB9
        {  "47",    "29",   "G7",  "21"  },                              // PB10
        {  "48",    "30",   "H7",  "22"  },                              // PB11
        {  "51",    "33",   "H8",  "25"  },                              // PB12
        {  "52",    "34",   "G8",  "26"  },                              // PB13
        {  "53",    "35",   "F8",  "27"  },                              // PB14
        {  "54",    "36",   "F7",  "28"  },                              // PB15
    },

    {
        //  LQFP100 LQFP64  BGA64  LQFP48                                GPIO C
        {  "15",    "8",    "E3",  "-"   },                              // PC0
        {  "16",    "9",    "E2",  "-"   },                              // PC1
        {  "17",    "10",   "F2",  "-"   },                              // PC2
        {  "18",    "11",   "-",   "-"   },                              // PC3
        {  "33",    "24",   "H5",  "-"   },                              // PC4
        {  "34",    "25",   "H6",  "-"   },                              // PC5
        {  "63",    "37",   "F6",  "-"   },                              // PC6
        {  "64",    "38",   "E7",  "-"   },                              // PC7
        {  "65",    "39",   "E8",  "-"   },                              // PC8
        {  "66",    "40",   "D8",  "-"   },                              // PC9
        {  "78",    "51",   "B7",  "-"   },                              // PC10
        {  "79",    "52",   "B6",  "-"   },                              // PC11
        {  "80",    "53",   "C5",  "-"   },                              // PC12
        {  "7",     "2",    "A2",  "2"   },                              // PC13
        {  "8",     "3",    "A1",  "3"   },                              // PC14
        {  "9",     "4",    "B1",  "4"   },                              // PC15
    },

    {
        //  LQFP100 LQFP64  BGA64  LQFP48                                GPIO D
        {  "81",    "5",    "C1",  "5"   },                              // PD0
        {  "82",    "6",    "D1",  "6"   },                              // PD1
        {  "83",    "54",   "B5",  "-"   },                              // PD2
        {  "84",    "-",    "-",   "-"   },                              // PD3
        {  "85",    "-",    "-",   "-"   },                              // PD4
        {  "86",    "-",    "-",   "-"   },                              // PD5
        {  "87",    "-",    "-",   "-"   },                              // PD6
        {  "88",    "-",    "-",   "-"   },                              // PD7
        {  "55",    "-",    "-",   "-"   },                              // PD8
        {  "56",    "-",    "-",   "-"   },                              // PD9
        {  "57",    "-",    "-",   "-"   },                              // PD10
        {  "58",    "-",    "-",   "-"   },                              // PD11
        {  "59",    "-",    "-",   "-"   },                              // PD12
        {  "60",    "-",    "-",   "-"   },                              // PD13
        {  "61",    "-",    "-",   "-"   },                              // PD14
        {  "62",    "-",    "-",   "-"   },                              // PD15
    },

    {
        //  LQFP100 LQFP64  BGA64  LQFP48                                GPIO E
        {  "97",    "-",    "-",   "-"   },                              // PE0
        {  "98",    "-",    "-",   "-"   },                              // PE1
        {  "1",     "-",    "-",   "-"   },                              // PE2
        {  "2",     "-",    "-",   "-"   },                              // PE3
        {  "3",     "-",    "-",   "-"   },                              // PE4
        {  "4",     "-",    "-",   "-"   },                              // PE5
        {  "5",     "-",    "-",   "-"   },                              // PE6
        {  "38",    "-",    "-",   "-"   },                              // PE7
        {  "39",    "-",    "-",   "-"   },                              // PE8
        {  "40",    "-",    "-",   "-"   },                              // PE9
        {  "41",    "-",    "-",   "-"   },                              // PE10
        {  "42",    "-",    "-",   "-"   },                              // PE11
        {  "43",    "-",    "-",   "-"   },                              // PE12
        {  "44",    "-",    "-",   "-"   },                              // PE13
        {  "45",    "-",    "-",   "-"   },                              // PE14
        {  "46",    "-",    "-",   "-"   },                              // PE15
    },
};